#include <assert.h>
#include "chain.h"

// ops are used later in reverse order
// op 0 means n |-> 2n
// op 1,2,3,5,7,...,W means n |-> n+op

std::vector<int> chain(mpz_class n,int W)
{
  assert(n >= 0);
  assert(W >= 1);
  assert(W % 2 == 1);
  std::vector<int> result;

  if (n == 0) {
    result.push_back(0);
    return result;
  }

  for (;;) {
    if (n <= W) {
      if (n % 2 != 0) break;
      if (n == 2 && W >= 3) break;
    }
    if (n == W+2 && W >= 3) {
      result.push_back(2);
      n = W;
      break;
    }
    if (W+4 <= n && n <= (3*W-2) && (n%6)==1) {
      result.push_back(mpz_class((n-4)/3).get_si());
      result.push_back(0);
      n = (n+2)/3;
      break;
    }
    if (W+4 <= n && n <= (3*W) && (n%6)==3) {
      result.push_back(mpz_class(n/3).get_si());
      result.push_back(0);
      n = n/3;
      break;
    }
    if (W+4 <= n && n <= (3*W-4) && (n%6)==5) {
      result.push_back(mpz_class((n+4)/3).get_si());
      result.push_back(0);
      n = (n-2)/3;
      break;
    }
    if (4 <= n && n <= (2*W-2) && (n%4)==0) {
      result.push_back(mpz_class(n/2-1).get_si());
      n = n/2+1;
      break;
    }
    if (n % 2 != 0) {
      int r,bits,bestr,bestbits;
      bestr = 1;
      bestbits = -1;
      for (r = 1;r <= W;r += 2) {
        bits = mpz_scan1(mpz_class(n - r).get_mpz_t(),0);
        if (bits > bestbits) {
          bestr = r;
          bestbits = bits;
        }
      }
      result.push_back(bestr);
      n -= bestr;
      continue;
    }
    result.push_back(0);
    n /= 2;
    continue;
  }

  result.push_back(n.get_si());

  return result;
}

static long long chain_cost(mpz_class n,int W)
{
  std::vector<int> ops = chain(n,W);
  long long result = 0;
  if (W >= 3) result += 4 + 5 * ((W-1)/2);
  for (long long i = 0;i < ops.size();++i)
    if (ops[i] == 0)
      result += 4;
    else
      result += 5;
  return result;
}

int chain_bestW(mpz_class n)
{
  int bestW = 1;
  long long bestcost = chain_cost(n,1);

  for (int W = 1;W <= 2 * mpz_sizeinbase(n.get_mpz_t(),2) + 3;W += 2) {
    long long cost = chain_cost(n,W);
    if (cost < bestcost) {
      bestW = W;
      bestcost = cost;
    }
  }

  return bestW;
}
