#include <iostream>
using namespace std;
#include "natural.h"
#include <assert.h>

void tinytest()
{
  natural x0(0,0);
  natural x1(1,1);
  natural x2(2,2);
  natural x3(3,2);

  cout << "x0 = " << x0.value() << '\n';
  cout << "x1 = " << x1.value() << '\n';
  cout << "x2 = " << x2.value() << '\n';
  cout << "x3 = " << x3.value() << '\n';

  cout << "x0+x0 = " << (x0+x0).value() << '\n';
  cout << "x0+x1 = " << (x0+x1).value() << '\n';
  cout << "x0+x2 = " << (x0+x2).value() << '\n';
  cout << "x0+x3 = " << (x0+x3).value() << '\n';
  cout << "x1+x0 = " << (x1+x0).value() << '\n';
  cout << "x1+x1 = " << (x1+x1).value() << '\n';
  cout << "x1+x2 = " << (x1+x2).value() << '\n';
  cout << "x1+x3 = " << (x1+x3).value() << '\n';
  cout << "x2+x0 = " << (x2+x0).value() << '\n';
  cout << "x2+x1 = " << (x2+x1).value() << '\n';
  cout << "x2+x2 = " << (x2+x2).value() << '\n';
  cout << "x2+x3 = " << (x2+x3).value() << '\n';
  cout << "x3+x0 = " << (x3+x0).value() << '\n';
  cout << "x3+x1 = " << (x3+x1).value() << '\n';
  cout << "x3+x2 = " << (x3+x2).value() << '\n';
  cout << "x3+x3 = " << (x3+x3).value() << '\n';

  cout << "ops " << bit::ops() << '\n';
}

void test(long long b0,mpz_class i0,long long b1,mpz_class i1)
{
  natural x0 = natural(i0,b0);
  natural x1 = natural(i1,b1);
  assert(x0.value() == natural(i0).value());
  assert(x0.value() == i0);
  assert((x0+x1).value() == i0+i1);
  assert((x0+x1).subtract(x1,b0).value() == i0);
  assert((x0+x1).subtract(i1,b0).value() == i0);
  if (i0 >= i1)
    assert(x0.subtract(x1,b0).value() == i0 - i1);
  else if (b0 < b1)
    assert(x0.subtract(x1,b1).value() == (mpz_class(1) << b1) + i0 - i1);
  else
    assert(x0.subtract(x1,b0).value() == (mpz_class(1) << b0) + i0 - i1);
  assert(x0.subtract(i1,b0).value() == x0.subtract(x1,b0).value());
  assert((x0+i1).value() == i0+i1);
  assert((x0*x1).value() == i0*i1);
  assert((x0*i1).value() == i0*i1);

  assert(x0.isequal(i1).value() == (i0==i1));

  if (i1 > 0) {
    assert((x0%i1).value() == i0%i1);

    if (!(random() & 127)) {
      mpz_class s = 1%i1;
      for (mpz_class e = 0;e < 100;++e) {
        assert(x0.powmod(e,i1).value() == s);
        assert(x0.powmod(natural(e),i1).value() == s);
        assert(x0.powpow8mod(natural(e),natural(3),i1).value() == (s*6561)%i1);
        s = (s*i0)%i1;
      }
    }

    if (mpz_probab_prime_p(i1.get_mpz_t(),3)) {
      mpz_class r = x0.invertmodprime(i1).value();
      if (i0 % i1 != 0)
        assert((r * i0) % i1 == 1);
      else
        if (i1 == 2)
          assert(r == 1);
        else
          assert(r == 0);
    }
  }
}

int main()
{
  tinytest();

  for (long long b0 = 0;b0 < 8;++b0)
    for (mpz_class i0 = 0;i0 < (mpz_class(1) << b0);++i0)
      for (long long b1 = 0;b1 < 8;++b1)
        for (mpz_class i1 = 0;i1 < (mpz_class(1) << b1);++i1)
          test(b0,i0,b1,i1);

  gmp_randstate_t gmprand;
  gmp_randinit_default(gmprand);

  for (long long b0 = 0;b0 < 256;++b0) {
    long long b1 = random() & 255;
    mpz_class i0;
    mpz_class i1;
    mpz_urandomb(i0.get_mpz_t(),gmprand,b0);
    mpz_urandomb(i1.get_mpz_t(),gmprand,b1);
    test(b0,i0,b1,i1);
  }

  return 0;
}
