#include <string.h>
#include "knownrandom.h"
#include "sodium/crypto_stream_chacha20.h"

#define crypto_rng_KEYBYTES 32
#define crypto_rng_OUTPUTBYTES 7648

#define crypto_stream crypto_stream_chacha20
#define KEYBYTES crypto_stream_chacha20_KEYBYTES
#define NONCEBYTES crypto_stream_chacha20_NONCEBYTES
#define OUTPUTBYTES crypto_rng_OUTPUTBYTES

#if KEYBYTES != crypto_rng_KEYBYTES
  KEYBYTES mismatch!
#endif

static const unsigned char nonce[NONCEBYTES] = {0};

static int crypto_rng(
        unsigned char *r, /* random output */
        unsigned char *n, /* new key */
  const unsigned char *g  /* old key */
)
{
  unsigned char x[KEYBYTES + OUTPUTBYTES];
  crypto_stream(x,sizeof x,nonce,g);
  memcpy(n,x,KEYBYTES);
  memcpy(r,x + KEYBYTES,OUTPUTBYTES);
  return 0;
}

static unsigned char g[crypto_rng_KEYBYTES];
static unsigned char r[crypto_rng_OUTPUTBYTES];
static unsigned long long pos = crypto_rng_OUTPUTBYTES;

void knownrandom_bytes(unsigned char *x,unsigned long long xlen)
{

#ifdef SIMPLE

  while (xlen > 0) {
    if (pos == crypto_rng_OUTPUTBYTES) {
      crypto_rng(r,g,g);
      pos = 0;
    }
    *x++ = r[pos]; xlen -= 1;
    r[pos++] = 0;
  }

#else /* same output but optimizing copies */

  while (xlen > 0) {
    unsigned long long ready;
    
    if (pos == crypto_rng_OUTPUTBYTES) {
      while (xlen > crypto_rng_OUTPUTBYTES) {
        crypto_rng(x,g,g);
        x += crypto_rng_OUTPUTBYTES;
        xlen -= crypto_rng_OUTPUTBYTES;
      }
      if (xlen == 0) return;
      
      crypto_rng(r,g,g);
      pos = 0;
    }
    
    ready = crypto_rng_OUTPUTBYTES - pos;
    if (xlen <= ready) ready = xlen;
    memcpy(x,r + pos,ready);
    memset(r + pos,0,ready);
    x += ready;
    xlen -= ready;
    pos += ready;
  }

#endif

}

unsigned int knownrandom_range(unsigned int u)
{
  long long top = (4294967296ULL / u) * u;
  unsigned int r;

  for (;;) {
    knownrandom_bytes((unsigned char *) &r,sizeof r);
    if (r < top) return r % u;
  }
}

unsigned int knownrandom_range3(void)
{
  unsigned int top = 4294967295ULL;
  unsigned int r;

  for (;;) {
    knownrandom_bytes((unsigned char *) &r,sizeof r);
    if (r < top) return r % 3;
  }
}
